import typing as t

import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.asyncio.processor.standard.visitor_selector import (  # noqa : E501
    select_rows,
)
from sarus_data_spec.path import path
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


async def filter_to_arrow(
    dataset: st.Dataset, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:

    previous_ds_id = dataset.protobuf().spec.transformed.arguments[0]
    previous_ds = t.cast(Dataset, dataset.storage().referrable(previous_ds_id))
    parent_schema = previous_ds.schema()
    data_type = sdt.Type(dataset.transform().protobuf().spec.filter.filter)
    data_type = update_fks(
        curr_type=data_type, original_type=parent_schema.type()  # type:ignore
    )

    async def async_generator(
        parent_iter: t.AsyncIterator[pa.RecordBatch],
    ) -> t.AsyncIterator[pa.RecordBatch]:
        async for batch in parent_iter:
            # TODO: what follows is really ugly and is necessary
            # because we do not have a proper type for the
            # protected entity, it will be removed when we
            # switch to that formalism
            array = convert_record_batch(
                record_batch=batch, _type=parent_schema.type()
            )
            if 'data' in previous_ds.schema().type().children():
                old_arrays = array.flatten()
                array = array.field('data')
                updated_array, filter_indices = select_rows(
                    dataset.schema().data_type(),
                    array,
                )
                old_arrays[0] = updated_array
                new_struct = pa.StructArray.from_arrays(
                    old_arrays,
                    names=list(previous_ds.schema().type().children().keys()),
                )
                yield pa.RecordBatch.from_struct_array(
                    new_struct.filter(filter_indices)
                )
            else:
                updated_array, filter_indices = select_rows(
                    dataset.schema().data_type(),
                    array,
                )
                yield pa.RecordBatch.from_struct_array(
                    updated_array.filter(filter_indices)
                )

    return async_generator(
        parent_iter=await previous_ds.async_to_arrow(batch_size=batch_size)
    )


def update_fks(curr_type: st.Type, original_type: st.Type) -> st.Type:
    class Select(st.TypeVisitor):
        result = curr_type

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
        ) -> None:
            new_fields = {}
            for fieldname, fieldtype in fields.items():
                new_fields[fieldname] = update_fks(
                    curr_type=fieldtype, original_type=original_type
                )

            self.result = sdt.Struct(
                fields=new_fields,
                name=name if name is not None else 'Struct',
                properties=curr_type.properties(),
            )
            # otherwise struct is empty and it is a terminal node

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
        ) -> None:
            new_fields = {}
            for fieldname, fieldtype in fields.items():
                new_fields[fieldname] = update_fks(
                    curr_type=fieldtype, original_type=original_type
                )

            self.result = sdt.Union(
                fields=new_fields,
                name=name if name is not None else 'Union',
                properties=curr_type.properties(),
            )

        def Optional(
            self, type: st.Type, name: t.Optional[str] = None
        ) -> None:

            self.result = sdt.Optional(
                type=update_fks(curr_type=type, original_type=original_type),
                name=name if name is not None else 'Optional',
                properties=curr_type.properties(),
            )

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
        ) -> None:
            self.result = sdt.List(
                type=update_fks(curr_type=type, original_type=original_type),
                max_size=max_size,
                name=name if name is not None else 'Optional',
                properties=curr_type.properties(),
            )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
        ) -> None:
            self.result = sdt.Array(
                type=update_fks(curr_type=type, original_type=original_type),
                shape=shape,
                name=name if name is not None else 'Optional',
                properties=curr_type.properties(),
            )

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
        ) -> None:
            if reference is not None:
                try:
                    original_type.get(path(paths=[reference]))
                except KeyError:
                    self.result = sdt.Id(unique=unique, base=base)

    visitor = Select()
    curr_type.accept(visitor)
    return visitor.result



import pyarrow

import sarus_data_spec.protobuf as sp
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


def from_arrow(type: pyarrow.DataType) -> sdt.Type:
    # Integers
    if pyarrow.types.is_int32(type):
        return sdt.Integer(base=st.IntegerBase.INT32)
    if pyarrow.types.is_int64(type):
        return sdt.Integer(base=st.IntegerBase.INT64)
    # Floats
    if pyarrow.types.is_float32(type):
        return sdt.Float(base=st.FloatBase.FLOAT32)
    if pyarrow.types.is_float64(type):
        return sdt.Float(base=st.FloatBase.FLOAT64)
    # Text
    if pyarrow.types.is_string(type):
        return sdt.Text()
    if pyarrow.types.is_boolean(type):
        return sdt.Boolean()
    if pyarrow.types.is_date32(type):
        return sdt.Datetime(base=st.DatetimeBase.INT64_NS, format='%Y-%m-%d')
    if pyarrow.types.is_timestamp(type):
        return sdt.Datetime(base=st.DatetimeBase.INT64_NS, format='%Y-%m-%d')
    if pyarrow.types.is_null(type):
        return sdt.Unit()
    raise NotImplementedError('Type not implemented')


def type_from_arrow(arrow_type: pyarrow.DataType, nullable: bool) -> st.Type:
    if nullable and not (pyarrow.types.is_null(arrow_type)):
        return sdt.Optional(type=from_arrow(arrow_type))
    return from_arrow(arrow_type)


def to_arrow(sarus_type: sp.Type, nullable: bool = False) -> pyarrow.DataType:
    """Convert Sarus type to PyArrow type.

    See https://arrow.apache.org/docs/python/api/datatypes.html.
    """
    # Integers
    if sarus_type.HasField("integer"):
        if sarus_type.integer.base == sp.Type.Integer.Base.INT64:
            return pyarrow.int64()
        elif sarus_type.integer.base == sp.Type.Integer.Base.INT32:
            return pyarrow.int32()
        elif sarus_type.integer.base == sp.Type.Integer.Base.INT16:
            return pyarrow.int16()
        elif sarus_type.integer.base == sp.Type.Integer.Base.INT8:
            return pyarrow.int8()

    # Enums
    if sarus_type.HasField("enum"):
        return pyarrow.string()

    # Floats
    elif sarus_type.HasField("float"):
        if sarus_type.float.base == sp.Type.Float.Base.FLOAT64:
            return pyarrow.float64()
        elif sarus_type.float.base == sp.Type.Float.Base.FLOAT32:
            return pyarrow.float32()
        elif sarus_type.float.base == sp.Type.Float.Base.FLOAT16:
            return pyarrow.float16()

    # Bytes
    elif sarus_type.HasField("bytes"):
        return pyarrow.binary()

    # Text
    elif sarus_type.HasField("text"):
        return pyarrow.string()
    # Id
    elif sarus_type.HasField("id"):

        if sarus_type.id.base == sp.Type.Id.Base.INT64:
            return pyarrow.int64()
        elif sarus_type.id.base == sp.Type.Id.Base.INT32:
            return pyarrow.int32()
        elif sarus_type.id.base == sp.Type.Id.Base.INT16:
            return pyarrow.int16()
        elif sarus_type.id.base == sp.Type.Id.Base.INT8:
            return pyarrow.int8()
        elif sarus_type.id.base == sp.Type.Id.Base.STRING:
            return pyarrow.string()
        else:
            return pyarrow.binary()

    # Boolean
    elif sarus_type.HasField("boolean"):
        return pyarrow.bool_()

    # Struct
    elif sarus_type.HasField("struct"):
        fields = sarus_type.struct.fields
        return pyarrow.struct(
            [
                pyarrow.field(
                    name=field.name,
                    type=to_arrow(field.type),
                    nullable=field.type.HasField('optional')
                    or field.type.HasField('unit'),
                )
                for field in fields
            ]
        )

    # Optional
    elif sarus_type.HasField("optional"):
        return to_arrow(sarus_type.optional.type, nullable=True)

    # List
    elif sarus_type.HasField("list"):
        return pyarrow.list_(
            to_arrow(sarus_type.list.type), sarus_type.list.max_size
        )

    # Union
    elif sarus_type.HasField("union"):
        return pyarrow.struct(
            [
                pyarrow.field(
                    name=field.name,
                    type=to_arrow(field.type),
                    nullable=True,
                )
                for field in sarus_type.union.fields
            ]
            + [pyarrow.field(name='field_selected', type=pyarrow.string())]
        )

    # Array
    elif sarus_type.HasField("array"):
        # No real mapping for array in PyArrow
        raise NotImplementedError

    elif sarus_type.HasField('datetime'):
        if sarus_type.datetime.base == sp.Type.Datetime.INT64_NS:
            return pyarrow.timestamp('ns')
        elif sarus_type.datetime.base == sp.Type.Datetime.INT64_MS:
            return pyarrow.timestamp('ms')
        else:
            return pyarrow.string()
    return pyarrow.null()



from typing import Any, AsyncIterator, cast
import json
import typing as t

import pyarrow as pa

from sarus_data_spec.constants import DATASET_SLUGNAME
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.path import Path, path
from sarus_data_spec.protobuf.typing import Protobuf
from sarus_data_spec.protobuf.utilities import from_base64, to_base64
from sarus_data_spec.schema import schema
import sarus_data_spec.protobuf as sp
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st

from .external.routing import arrow_external, external
from .standard.extract import arrow_extract
from .standard.filter import filter_to_arrow
from .standard.get_item import get_item_to_arrow
from .standard.project import project_to_arrow
from .standard.sample import arrow_sample
from .standard.select_sql import select_sql_schema, select_sql_to_arrow
from .standard.slice import arrow_slice


async def transformed_dataset_arrow(
    dataset: st.Dataset, batch_size: int
) -> AsyncIterator[pa.RecordBatch]:
    """Routes a transformed Dataspec to its Arrow implementation."""
    transform = dataset.transform()
    if transform.is_external():
        return await arrow_external(dataset, batch_size)
    elif transform.protobuf().spec.HasField('sample'):
        return await arrow_sample(dataset, batch_size)
    elif transform.protobuf().spec.HasField('filter'):
        return await filter_to_arrow(dataset, batch_size)
    elif transform.protobuf().spec.HasField('project'):
        return await project_to_arrow(dataset, batch_size)
    elif transform.protobuf().spec.HasField('get_item'):
        return await get_item_to_arrow(dataset, batch_size)
    elif transform.protobuf().spec.HasField('select_sql'):
        return await select_sql_to_arrow(
            dataset, transform.protobuf().spec.select_sql.query, batch_size
        )
    elif transform.protobuf().spec.HasField('extract'):
        return await arrow_extract(dataset, batch_size)
    elif transform.protobuf().spec.HasField('slice'):
        return await arrow_slice(dataset, batch_size)
    else:
        raise NotImplementedError(
            f"{transform.protobuf().spec.WhichOneof('spec')}"
        )


async def transformed_scalar(scalar: st.Scalar) -> Any:
    """Routes a transformed Scalar to its implementation.

    For now, transformed Scalars are only obtained from external ops. By
    external ops, we mean functions defined in external libraries.
    """
    if scalar.transform().is_external():
        return await external(scalar)
    else:
        raise NotImplementedError(
            f"scalar_transformed for {scalar.transform()}"
        )


def transformed_schema(dataset: st.Dataset) -> st.Schema:
    previous_ds_id = dataset.protobuf().spec.transformed.arguments[
        0
    ]  # how this works?
    previous_schema = cast(
        Dataset, dataset.storage().referrable(previous_ds_id)
    ).schema()
    if dataset.transform().protobuf().spec.HasField('filter'):
        new_type = sdt.Type(dataset.transform().protobuf().spec.filter.filter)
        new_type = update_fks(
            curr_type=new_type, original_type=new_type  # type:ignore
        )
        old_properties = previous_schema.properties()

        if 'primary_keys' in old_properties.keys():
            new_pks = filter_primary_keys(
                old_properties['primary_keys'],
                new_type,
            )
            old_properties['primary_keys'] = new_pks  # type:ignore

        # VERY UGLY SHOULD BE REMOVED WHEN WE HAVE PROTECTED TYPE
        previous_fields = previous_schema.type().children()
        if 'data' in previous_fields.keys():
            previous_fields['data'] = new_type
            new_type = sdt.Struct(fields=previous_fields)
        return schema(
            dataset,
            schema_type=new_type,
            protected_paths=previous_schema.protobuf().protected,
            properties=old_properties,
            name=dataset.properties().get(DATASET_SLUGNAME, None),
        )

    if dataset.transform().protobuf().spec.HasField('get_item'):
        path = Path(dataset.transform().protobuf().spec.get_item.path)
        sub_types = previous_schema.data_type().sub_types(path)
        assert len(sub_types) == 1
        new_type = sub_types[0]  # type:ignore
        # TODO: update foreign_keys/primary_keys in the type
        previous_fields = previous_schema.type().children()
        if 'data' in previous_fields.keys():
            previous_fields['data'] = new_type
            new_type = sdt.Struct(fields=previous_fields)
        return schema(
            dataset,
            schema_type=new_type,
            protected_paths=previous_schema.protobuf().protected,
            name=dataset.properties().get(DATASET_SLUGNAME, None),
        )

    elif (
        dataset.transform().protobuf().spec.HasField('shuffle')
        or dataset.transform().protobuf().spec.HasField('sample')
        or dataset.transform().protobuf().spec.HasField('slice')
        or dataset.transform().protobuf().spec.HasField('extract')
    ):
        return schema(
            dataset,
            schema_type=sdt.Type(previous_schema.protobuf().type),
            protected_paths=previous_schema.protobuf().protected,
            properties=previous_schema.properties(),
            name=dataset.properties().get(DATASET_SLUGNAME, None),
        )

    elif dataset.transform().protobuf().spec.HasField('select_sql'):
        parent_ds = cast(Dataset, dataset.storage().referrable(previous_ds_id))
        new_type = select_sql_schema(
            parent_ds,
            dataset.transform().protobuf().spec.select_sql.query,
        )
        return schema(
            dataset,
            schema_type=new_type,
        )

    raise ValueError('Other spec not implemented yet')


def filter_primary_keys(old_pks: str, new_type: st.Type) -> str:
    """Keeps only primary keys path that appear in new type"""
    filtered_pks = []
    primary_keys = [
        Path(cast(sp.Path, from_base64(proto, cast(Protobuf, sp.Path))))
        for proto in json.loads(old_pks)
    ]
    for primary_key in primary_keys:
        try:
            new_type.get(primary_key)
        except KeyError:
            pass

        else:
            filtered_pks.append(to_base64(primary_key.protobuf()))
    return json.dumps(filtered_pks)


def update_fks(curr_type: st.Type, original_type: st.Type) -> st.Type:
    class Select(st.TypeVisitor):
        result = curr_type

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
        ) -> None:
            new_fields = {}
            for fieldname, fieldtype in fields.items():
                new_fields[fieldname] = update_fks(
                    curr_type=fieldtype, original_type=original_type
                )

            self.result = sdt.Struct(
                fields=new_fields,
                name=name if name is not None else 'Struct',
                properties=curr_type.properties(),
            )
            # otherwise struct is empty and it is a terminal node

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
        ) -> None:
            new_fields = {}
            for fieldname, fieldtype in fields.items():
                new_fields[fieldname] = update_fks(
                    curr_type=fieldtype, original_type=original_type
                )

            self.result = sdt.Union(
                fields=new_fields,
                name=name if name is not None else 'Union',
                properties=curr_type.properties(),
            )

        def Optional(
            self, type: st.Type, name: t.Optional[str] = None
        ) -> None:

            self.result = sdt.Optional(
                type=update_fks(curr_type=type, original_type=original_type),
                name=name if name is not None else 'Optional',
                properties=curr_type.properties(),
            )

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
        ) -> None:
            self.result = sdt.List(
                type=update_fks(curr_type=type, original_type=original_type),
                max_size=max_size,
                name=name if name is not None else 'Optional',
                properties=curr_type.properties(),
            )

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
        ) -> None:
            self.result = sdt.Array(
                type=update_fks(curr_type=type, original_type=original_type),
                shape=shape,
                name=name if name is not None else 'Optional',
                properties=curr_type.properties(),
            )

        def Id(
            self,
            unique: bool,
            reference: t.Optional[st.Path] = None,
            base: t.Optional[st.IdBase] = None,
        ) -> None:
            if reference is not None:
                try:
                    original_type.get(path(paths=[reference]))
                except KeyError:
                    self.result = sdt.Id(unique=unique, base=base)

    visitor = Select()
    curr_type.accept(visitor)
    return visitor.result



import typing

import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.asyncio.processor.standard.visitor_selector import (  # noqa : E501
    select_rows,
)
from sarus_data_spec.path import Path
import sarus_data_spec.typing as st


async def get_item_to_arrow(
    dataset: st.Dataset, batch_size: int
) -> typing.AsyncIterator[pa.RecordBatch]:

    previous_ds_id = dataset.protobuf().spec.transformed.arguments[0]
    previous_ds = typing.cast(
        Dataset, dataset.storage().referrable(previous_ds_id)
    )
    path = Path(dataset.transform().protobuf().spec.get_item.path)
    parent_schema = previous_ds.schema()

    async def async_generator(
        parent_iter: typing.AsyncIterator[pa.RecordBatch],
    ) -> typing.AsyncIterator[pa.RecordBatch]:
        async for batch in parent_iter:
            # TODO: what follows is really ugly and is necessary
            # because we do not have a proper type for the
            # protected entity, it will be removed when we
            # switch to that formalism
            array = convert_record_batch(
                record_batch=batch, _type=parent_schema.type()
            )
            array = convert_record_batch(
                record_batch=batch, _type=previous_ds.schema().type()
            )
            # VERY UGLY SHOULD BE REMOVED WHEN WE HAVE PROTECTED TYPE
            if 'data' in previous_ds.schema().type().children():
                old_arrays = array.flatten()
                array = array.field('data')
                updated_array = get_items(
                    _type=previous_ds.schema().data_type(),
                    array=array,
                    path=path,
                )
                old_arrays[0] = updated_array
                new_struct = pa.StructArray.from_arrays(
                    old_arrays,
                    names=list(previous_ds.schema().type().children().keys()),
                )
                yield pa.RecordBatch.from_struct_array(new_struct)
            else:
                updated_array = get_items(
                    _type=previous_ds.schema().data_type(),
                    array=array,
                    path=path,
                )
                if isinstance(updated_array, pa.StructArray):
                    yield pa.RecordBatch.from_struct_array(updated_array)
                else:
                    yield pa.RecordBatch.from_arrays(
                        [updated_array], names=[path.to_strings_list()[0][-1]]
                    )

    return async_generator(
        parent_iter=await previous_ds.async_to_arrow(batch_size=batch_size)
    )


def get_items(array: pa.Array, path: st.Path, _type: st.Type) -> pa.Array:
    """Visitor selecting columns based on the type.
    The idea is that at each level,
    the filter for the array is computed, and for the union,
    we remove the fields that we want to filter among
    the columns
    """

    class ItemSelector(st.TypeVisitor):
        batch_array: pa.Array = array

        def Struct(
            self,
            fields: typing.Mapping[str, st.Type],
            name: typing.Optional[str] = None,
        ) -> None:

            if len(path.sub_paths()) > 0:
                sub_path = path.sub_paths()[0]
                self.batch_array = get_items(
                    array=array.field(sub_path.label()),
                    path=sub_path,
                    _type=fields[sub_path.label()],
                )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: typing.Optional[str] = None,
        ) -> None:
            raise NotImplementedError

        def Optional(
            self, type: st.Type, name: typing.Optional[str] = None
        ) -> None:
            array = self.batch_array.field(path.label())
            if len(path.sub_paths()) == 0:
                self.batch_array = array
            else:
                self.batch_array = get_items(
                    array=array, path=path.sub_paths()[0], _type=type
                )

        def Union(
            self,
            fields: typing.Mapping[str, st.Type],
            name: typing.Optional[str] = None,
        ) -> None:

            if len(path.sub_paths()) == 0:
                self.batch_array = array
            else:
                sub_path = path.sub_paths()[0]
                self.batch_array = get_items(
                    array=array.field(sub_path.label()),
                    path=sub_path,
                    _type=fields[sub_path.label()],
                )

        def Array(
            self,
            type: st.Type,
            shape: typing.Tuple[int, ...],
            name: typing.Optional[str] = None,
        ) -> None:
            raise NotImplementedError

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: typing.Optional[str] = None,
        ) -> None:
            raise NotImplementedError

    visitor = ItemSelector()
    _type.accept(visitor)
    return visitor.batch_array



import typing as t

import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.ops.asyncio.processor.standard.filter import (  # noqa : E501
    update_fks,
)
from sarus_data_spec.manager.ops.asyncio.processor.standard.visitor_selector import (  # noqa : E501
    select_columns,
)
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


async def project_to_arrow(
    dataset: st.Dataset, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:

    previous_ds_id = dataset.protobuf().spec.transformed.arguments[0]
    previous_ds = t.cast(Dataset, dataset.storage().referrable(previous_ds_id))

    parent_schema = previous_ds.schema()
    data_type = sdt.Type(
        dataset.transform().protobuf().spec.project.projection
    )
    data_type = update_fks(
        curr_type=data_type, original_type=parent_schema.type()  # type:ignore
    )

    async def async_generator(
        parent_iter: t.AsyncIterator[pa.RecordBatch],
    ) -> t.AsyncIterator[pa.RecordBatch]:
        async for batch in parent_iter:
            updated_array = select_columns(
                data_type,
                convert_record_batch(
                    record_batch=batch, _type=parent_schema.type()
                ),
            )
            yield pa.RecordBatch.from_struct_array(updated_array)

    return async_generator(
        parent_iter=await previous_ds.async_to_arrow(batch_size=batch_size)
    )



import typing as t

import pyarrow as pa

from sarus_data_spec.arrow.type import type_from_arrow
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.manager.asyncio.utils import async_iter
import sarus_data_spec.type as sdt
import sarus_data_spec.typing as st


async def select_sql_to_arrow(
    dataset: st.Dataset, query: str, batch_size: int
) -> t.AsyncIterator[pa.RecordBatch]:
    """
    Pyarrow batches from results of an SQL query
    executed on the parent dataset
    """
    # TODO: Adapt when dataset.sql will return an arrow iterator with
    # the schema inferred from sqlalchemy results.
    parent_ds_id = dataset.protobuf().spec.transformed.arguments[0]
    parent_ds = t.cast(Dataset, dataset.storage().referrable(parent_ds_id))
    res = pa.Table.from_pylist(parent_ds.sql(query))
    if len(res) == 0:
        raise RuntimeError(
            f"SQL query: '{query}' without any result is not supported."
        )
    batches = res.to_batches(batch_size)
    return async_iter(batches)


def select_sql_schema(dataset: st.Dataset, query: str) -> sdt.Type:
    """
    Schema computed by inferring results of an SQL query
    executed on the parent dataset
    """
    # TODO: Adapt when dataset.sql will return an arrow iterator with
    # the schema inferred from sqlalchemy results.
    res_table = pa.Table.from_pylist(dataset.sql(query))
    if len(res_table) == 0:
        raise RuntimeError(
            f"SQL query: '{query}' without any result is not supported."
        )
    fields = {
        col: type_from_arrow(
            arrow_type=res_table.field(col).type,
            nullable=res_table.field(col).nullable,
        )
        for col in res_table.schema.names
    }

    return sdt.Struct(fields=fields)



from typing import AsyncIterator, cast

import pyarrow as pa

from sarus_data_spec.manager.asyncio.utils import async_iter
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


async def arrow_extract(
    dataset: st.Dataset, batch_size: int
) -> AsyncIterator[pa.RecordBatch]:
    """Implementation of the extract transform
    At the moment is identical to slice but in the future it should it
    should be randomized
    """
    args, kwargs = dataset.parents()
    parents = args + list(kwargs.values())
    assert len(parents) == 1
    (parent,) = parents
    assert parent.prototype() == sp.Dataset
    ds_parent = cast(st.Dataset, parent)
    async_parent_batches = await ds_parent.async_to_arrow()
    parent_batches = [batch async for batch in async_parent_batches]
    parent_table = pa.Table.from_batches(parent_batches)
    size = dataset.transform().protobuf().spec.extract.size
    if size < 0:
        raise ValueError("Extract transform size parameter must be positive")
    batches = (
        parent_table.slice(offset=0, length=size)
        .combine_chunks()
        .to_batches(batch_size)
    )
    return async_iter(batches)



from __future__ import annotations

import datetime
import json
import typing as t

import numpy as np
import pandas as pd
import pyarrow as pa

from sarus_data_spec.base import Base
from sarus_data_spec.constants import (
    ARRAY_VALUES,
    LIST_VALUES,
    OPTIONAL_VALUE,
    TEXT_CHARSET,
    TEXT_MAX_LENGTH,
)
from sarus_data_spec.path import Path
from sarus_data_spec.predicate import Predicate
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as sdt

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass


class Type(Base[sp.Type]):
    """A python class to describe types"""

    def prototype(self) -> t.Type[sp.Type]:
        """Return the type of the underlying protobuf."""
        return sp.Type

    def name(self) -> str:
        """Returns the name of the underlying protobuf."""
        return self.protobuf().name

    # A Visitor acceptor
    def accept(self, visitor: sdt.TypeVisitor) -> None:
        dispatch: t.Callable[[], None] = {
            'null': visitor.Null,
            'unit': visitor.Unit,
            'boolean': visitor.Boolean,
            'integer': lambda: visitor.Integer(
                min=self._protobuf.integer.min,
                max=self._protobuf.integer.max,
                base=sdt.IntegerBase(self._protobuf.integer.base),
            ),
            'id': lambda: visitor.Id(
                base=sdt.IdBase(self._protobuf.id.base),
                unique=self._protobuf.id.unique,
                reference=Path(self._protobuf.id.reference)
                if self._protobuf.id.reference != sp.Path()
                else None,
            ),
            'enum': lambda: visitor.Enum(
                self._protobuf.name,
                [
                    (name_value.name, name_value.value)
                    for name_value in self._protobuf.enum.name_values
                ],
                self._protobuf.enum.ordered,
            ),
            'float': lambda: visitor.Float(
                min=self._protobuf.float.min,
                max=self._protobuf.float.max,
                base=sdt.FloatBase(self._protobuf.float.base),
            ),
            'text': lambda: visitor.Text(self._protobuf.text.encoding),
            'bytes': visitor.Bytes,
            'struct': lambda: visitor.Struct(
                {
                    field.name: Type(field.type)
                    for field in self._protobuf.struct.fields
                },
                name=None
                if self._protobuf.name == ''
                else self._protobuf.name,
            ),
            'union': lambda: visitor.Union(
                {
                    field.name: Type(field.type)
                    for field in self._protobuf.union.fields
                },
                name=None
                if self._protobuf.name == ''
                else self._protobuf.name,
            ),
            'optional': lambda: visitor.Optional(
                Type(self._protobuf.optional.type),
                None if self._protobuf.name == '' else self._protobuf.name,
            ),
            'list': lambda: visitor.List(
                Type(self._protobuf.list.type),
                max_size=self._protobuf.list.max_size,
                name=None
                if self._protobuf.name == ''
                else self._protobuf.name,
            ),
            'array': lambda: visitor.Array(
                Type(self._protobuf.array.type),
                tuple(self._protobuf.array.shape),
                None if self._protobuf.name == '' else self._protobuf.name,
            ),
            'datetime': lambda: visitor.Datetime(
                self._protobuf.datetime.format,
                self._protobuf.datetime.min,
                self._protobuf.datetime.max,
                sdt.DatetimeBase(self._protobuf.datetime.base),
            ),
            'constrained': lambda: visitor.Constrained(
                Type(self._protobuf.constrained.type),
                Predicate(self._protobuf.constrained.constraint),
                None if self._protobuf.name == '' else self._protobuf.name,
            ),
            'hypothesis': lambda: visitor.Hypothesis(
                *[
                    (Type(scored.type), scored.score)
                    for scored in self._protobuf.hypothesis.types
                ],
                name=None
                if self._protobuf.name == ''
                else self._protobuf.name,
            ),
            None: lambda: None,
        }[self._protobuf.WhichOneof('type')]
        dispatch()

    def latex(self: sdt.Type, parenthesized: bool = False) -> str:
        """return a latex representation of the type"""

        class Latex(sdt.TypeVisitor):
            result: str = ''

            def Null(self) -> None:
                self.result = r'\emptyset'

            def Unit(self) -> None:
                self.result = r'\mathbb{1}'

            def Boolean(self) -> None:
                self.result = r'\left\{0,1\right}'

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                if (
                    min <= np.iinfo(np.int32).min
                    or max >= np.iinfo(np.int32).max
                ):
                    self.result = r'\mathbb{N}'
                else:
                    self.result = (
                        r'\left[' + str(min) + r'..' + str(max) + r'\right]'
                    )

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                if len(name_values) > 3:
                    self.result = r'\left\{'
                    for name, _ in name_values[:2]:
                        self.result += r'\text{' + name + r'}, '
                    self.result += r',\ldots, '
                    for name, _ in name_values[-1:]:
                        self.result += r'\text{' + name + r'}, '
                    self.result = self.result[:-2] + r'\right\}'
                elif len(name_values) > 0:
                    self.result = r'\left\{'
                    for name, _ in name_values:
                        self.result += r'\text{' + name + r'}, '
                    self.result = self.result[:-2] + r'\right\}'
                else:
                    self.Unit()

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                if (
                    min <= np.finfo(np.float32).min
                    or max >= np.finfo(np.float32).max
                ):
                    self.result = r'\mathbb{R}'
                else:
                    self.result = (
                        r'\left[' + str(min) + r', ' + str(max) + r'\right]'
                    )

            def Text(self, encoding: str) -> None:
                self.result = r'\text{Text}'

            def Bytes(self) -> None:
                self.result = r'\text{Bytes}'

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                if len(fields) > 0:
                    if name is None:
                        self.result = r'\left\{'
                    else:
                        self.result = r'\text{' + name + r'}: \left\{'
                    for type_name, type in fields.items():
                        self.result = (
                            self.result
                            + r'\text{'
                            + type_name
                            + r'}:'
                            + Type.latex(type, parenthesized=True)
                            + r', '
                        )
                    self.result = self.result[:-2] + r'\right\}'
                else:
                    self.Unit()

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                if len(fields) > 0:
                    for type in fields.values():
                        self.result = (
                            self.result
                            + Type.latex(type, parenthesized=True)
                            + r' | '
                        )
                    self.result = self.result[:-2]
                    if parenthesized:
                        self.result = r'\left(' + self.result + r'\right)'
                else:
                    self.Null()

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = Type.latex(type, parenthesized=True) + r'?'

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                if max_size < 100:
                    self.result = (
                        Type.latex(type, parenthesized=True)
                        + r'^{'
                        + str(max_size)
                        + r'}'
                    )
                else:
                    self.result = Type.latex(type, parenthesized=True) + r'^*'

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = (
                    Type.latex(type)
                    + r'^{'
                    + r'\times '.join([str(i) for i in shape])
                    + r'}'
                )

            def Datetime(
                self,
                format: str,
                min: str,
                max: str,
                base: sdt.DatetimeBase,
            ) -> None:
                self.result = r'\text{Date}'

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:
                if len(types) > 0:
                    for type, score in types:
                        self.result = (
                            self.result
                            + Type.latex(type, parenthesized=True)
                            + f',{score}|'
                        )
                    self.result = self.result[:-2]
                    self.result = r'\langle' + self.result + r'\rangle'
                else:
                    self.Null()

        visitor = Latex()
        self.accept(visitor)
        return visitor.result

    def compact(self: sdt.Type, parenthesized: bool = False) -> str:
        """return a compact representation of the type"""

        class Compact(sdt.TypeVisitor):
            result: str = ''

            def Null(self) -> None:
                self.result = r'∅'
                self.result = r'∅'

            def Unit(self) -> None:
                self.result = r'𝟙'

            def Boolean(self) -> None:
                self.result = r'𝔹'

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                if (
                    min <= np.iinfo(np.int32).min
                    or max >= np.iinfo(np.int32).max
                ):
                    self.result = r'ℕ'
                else:
                    self.result = r'[' + str(min) + r'..' + str(max) + r']'

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                if len(name_values) > 3:
                    self.result = r'{'
                    for name, _ in name_values[:2]:
                        self.result += name
                    self.result += r',..., '
                    for name, _ in name_values[-1:]:
                        self.result += name + r', '
                    self.result = self.result[:-2] + r'}'
                elif len(name_values) > 0:
                    self.result = r'{'
                    for name, _ in name_values:
                        self.result += name + r', '
                    self.result = self.result[:-2] + r'}'
                else:
                    self.Unit()

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                if (
                    min <= np.finfo(np.float32).min
                    or max >= np.finfo(np.float32).max
                ):
                    self.result = r'ℝ'
                else:
                    self.result = r'[' + str(min) + r', ' + str(max) + r']'

            def Text(self, encoding: str) -> None:
                self.result = r'𝒯'

            def Bytes(self) -> None:
                self.result = r'ℬ'

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                if len(fields) > 0:
                    if name is None:
                        self.result = '{'
                    else:
                        self.result = name + r': {'
                    for type_name, type in fields.items():
                        self.result = (
                            self.result
                            + type_name
                            + r': '
                            + Type.compact(type, parenthesized=True)
                            + r', '
                        )
                    self.result = self.result[:-2] + r'}'
                else:
                    self.Unit()

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                if len(fields) > 0:
                    for type in fields.values():
                        self.result = (
                            self.result
                            + Type.compact(type, parenthesized=True)
                            + r' | '
                        )
                    self.result = self.result[:-2]
                    if parenthesized:
                        self.result = r'(' + self.result + r')'
                else:
                    self.Null()

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = Type.compact(type, parenthesized=True) + r'?'

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                self.result = Type.compact(type, parenthesized=True) + r'*'

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = (
                    Type.compact(type)
                    + r'**('
                    + r'x'.join([str(i) for i in shape])
                    + r')'
                )

            def Datetime(
                self,
                format: str,
                min: str,
                max: str,
                base: sdt.DatetimeBase,
            ) -> None:
                self.result = r'𝒟'

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:
                if len(types) > 0:
                    self.result = r'<'
                    for type, score in types:
                        self.result = (
                            self.result
                            + Type.compact(type, parenthesized=False)
                            + f',{score}|'
                        )
                    self.result = self.result[:-1] + r'>'
                else:
                    self.Null()

        visitor = Compact()
        self.accept(visitor)
        return visitor.result

    def get(self: Type, item: sdt.Path) -> sdt.Type:
        """Return a projection of the considered type defined by the path.
        The projection contains all the parents types of the leaves of
        the path. If the path stops at a Union, Struct or Optional,
        it also returns that type with everything it contains."""

        class Select(sdt.TypeVisitor):
            result = Type(sp.Type())

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties

            def Null(self) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Null()

            def Id(
                self,
                unique: bool,
                reference: t.Optional[sdt.Path] = None,
                base: t.Optional[sdt.IdBase] = None,
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Id(base=base, unique=unique, reference=reference)

            def Unit(self) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Unit(properties=self.properties)

            def Boolean(self) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Boolean()

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Integer(
                    min=min, max=max, properties=self.properties
                )

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Enum(
                    name=name,
                    name_values=name_values,
                    ordered=ordered,
                    properties=self.properties,
                )

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Float(
                    min=min, max=max, properties=self.properties
                )

            def Text(self, encoding: str) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Text(
                    encoding=encoding, properties=self.properties
                )

            def Bytes(self) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Bytes()

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                proto = item.protobuf()
                new_fields = {}
                for path in proto.paths:
                    # here struct each path must have a label
                    new_fields[path.label] = fields[path.label].get(Path(path))
                self.result = Struct(
                    fields=new_fields if len(new_fields) > 0 else fields,
                    name=name if name is not None else 'Struct',
                    properties=self.properties,
                )

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                proto = item.protobuf()
                new_fields = {}
                for path in proto.paths:
                    new_fields[path.label] = fields[path.label].get(Path(path))
                self.result = Union(
                    fields=new_fields if len(new_fields) > 0 else fields,
                    name=name if name is not None else 'Union',
                    properties=self.properties,
                )

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:

                proto = item.protobuf()
                assert len(proto.paths) <= 1
                self.result = Optional(
                    type.get(Path(proto.paths[0]))
                    if len(proto.paths) > 0
                    else type,
                    name=t.cast(str, name),
                    properties=self.properties,
                )

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) <= 1
                self.result = List(
                    type.get(Path(proto.paths[0]))
                    if len(proto.paths) > 0
                    else type,
                    name=t.cast(str, name),
                    max_size=max_size,
                    properties=self.properties,
                )

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) <= 1
                self.result = Array(
                    type.get(Path(proto.paths[0]))
                    if len(proto.paths) > 0
                    else type,
                    name=t.cast(str, name),
                    shape=shape,
                    properties=self.properties,
                )

            def Datetime(
                self,
                format: str,
                min: str,
                max: str,
                base: sdt.DatetimeBase,
            ) -> None:
                proto = item.protobuf()
                assert len(proto.paths) == 0
                self.result = Datetime(
                    format=format,
                    min=min,
                    max=max,
                    properties=self.properties,
                    base=base,
                )

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:

                # TODO
                pass

        visitor = Select(properties=self.properties())
        self.accept(visitor)
        return visitor.result

    def sub_types(self: Type, item: sdt.Path) -> t.List[sdt.Type]:
        """Returns a list of the subtypes corresponding to the
        leaves of the input path"""

        class Select(sdt.TypeVisitor):
            def __init__(self, type_item: sdt.Type):
                self.result = [type_item]

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                for sub_path in item.sub_paths():
                    result.extend(fields[sub_path.label()].sub_types(sub_path))
                if len(result) > 0:
                    self.result = result
                    # otherwise struct is empty and it is a terminal node

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                for sub_path in item.sub_paths():
                    result.extend(fields[sub_path.label()].sub_types(sub_path))
                if len(result) > 0:
                    self.result = result
                    # otherwise union is empty and it is a terminal node

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:

                result = []
                if len(item.sub_paths()) == 1:
                    result.extend(type.sub_types(item.sub_paths()[0]))
                    self.result = result

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                if len(item.sub_paths()) == 1:
                    result.extend(type.sub_types(item.sub_paths()[0]))
                    self.result = result

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                if len(item.sub_paths()) == 1:
                    result.extend(type.sub_types(item.sub_paths()[0]))
                    self.result = result

        visitor = Select(type_item=self)
        self.accept(visitor)
        return visitor.result

    def structs(self: Type) -> t.Optional[t.List[sdt.Path]]:
        """Returns the path to the first level structs encountered in the type.
        For example, Union[Struct1,Union[Struct2[Struct3]] will return only a
        path that brings to Struct1 and Struct2.
        """

        class AddPath(sdt.TypeVisitor):
            result: t.Optional[t.List[sdt.Path]] = None

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = []

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                paths = []
                for type_name, curr_type in fields.items():
                    if curr_type.protobuf().WhichOneof('type') == 'struct':
                        paths.append(Path(sp.Path(label=type_name)))
                    else:
                        sub_paths = curr_type.structs()
                        if sub_paths is not None:
                            paths.extend(
                                [
                                    Path(
                                        sp.Path(
                                            label=type_name,
                                            paths=[subpath.protobuf()],
                                        )
                                    )
                                    for subpath in sub_paths
                                ]
                            )
                if len(paths) > 0:
                    self.result = t.cast(t.List[sdt.Path], paths)

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                if type.protobuf().WhichOneof('type') == 'struct':
                    self.result = [Path(sp.Path(label=OPTIONAL_VALUE))]
                else:
                    sub_paths = type.structs()
                    if sub_paths is not None:
                        self.result = [
                            Path(
                                sp.Path(
                                    label=OPTIONAL_VALUE,
                                    paths=[
                                        subpath.protobuf()
                                        for subpath in sub_paths
                                    ],
                                )
                            )
                        ]

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                if type.protobuf().WhichOneof('type') == 'struct':
                    self.result = [Path(sp.Path(label=LIST_VALUES))]
                else:
                    sub_paths = type.structs()
                    if sub_paths is not None:
                        self.result = [
                            Path(
                                sp.Path(
                                    label=LIST_VALUES,
                                    paths=[
                                        subpath.protobuf()
                                        for subpath in sub_paths
                                    ],
                                )
                            )
                        ]

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                if type.protobuf().WhichOneof('type') == 'struct':
                    self.result = [Path(sp.Path(label=ARRAY_VALUES))]
                else:
                    sub_paths = type.structs()
                    if sub_paths is not None:
                        self.result = [
                            Path(
                                sp.Path(
                                    label=ARRAY_VALUES,
                                    paths=[
                                        subpath.protobuf()
                                        for subpath in sub_paths
                                    ],
                                )
                            )
                        ]

        visitor = AddPath()
        self.accept(visitor)
        return visitor.result

    def leaves(self: sdt.Type) -> t.List[sdt.Type]:
        """Returns a list of the sub-types corresponding to
        the leaves of the type tree structure"""

        class AddLeaves(sdt.TypeVisitor):
            result: t.List[sdt.Type] = []

            def __init__(self, type_item: sdt.Type):
                self.result = [type_item]

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                for item_name in fields.keys():
                    result.extend(fields[item_name].leaves())
                if len(result) > 0:
                    self.result = result

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                for item_name in fields.keys():
                    result.extend(fields[item_name].leaves())
                if len(result) > 0:
                    self.result = result

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:

                result = []
                result.extend(type.leaves())
                if len(result) > 0:
                    self.result = result

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                result.extend(type.leaves())
                if len(result) > 0:
                    self.result = result

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                result = []
                result.extend(type.leaves())
                if len(result) > 0:
                    self.result = result

        visitor = AddLeaves(type_item=self)
        self.accept(visitor)
        return visitor.result

    def children(self: sdt.Type) -> t.Dict[str, sdt.Type]:
        """Returns the children contained in the type tree structure"""

        class GetChildren(sdt.TypeVisitor):
            result: t.Dict[str, sdt.Type] = {}

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = t.cast(t.Dict[str, sdt.Type], fields)

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = t.cast(t.Dict[str, sdt.Type], fields)

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = {OPTIONAL_VALUE: type}

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                self.result = {LIST_VALUES: type}

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = {ARRAY_VALUES: type}

        visitor = GetChildren(properties=self.properties())
        self.accept(visitor)
        return visitor.result

    def example(self) -> pa.Array:
        """This methods returns a pyarrow scalar that matches the type.
        For an optional type, we consider the case where, the field
        is not missing.
        """

        class ToArrow(sdt.TypeVisitor):
            result = pa.nulls(0)

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties if properties is not None else {}

            def Null(self) -> None:
                raise NotImplementedError

            def Unit(self) -> None:
                self.result = pa.nulls(0)

            def Boolean(self) -> None:
                self.result = pa.array([True], type=pa.bool_())

            def Id(
                self,
                unique: bool,
                reference: t.Optional[sdt.Path] = None,
                base: t.Optional[sdt.IdBase] = None,
            ) -> None:
                # TODO: we should clarify for Ids, user_input
                # and so on, to be consistent
                if base == sdt.IdBase.STRING:
                    self.result = pa.array(['1'], pa.string())
                elif base == sdt.IdBase.INT64:
                    self.result = pa.array([1], pa.int64())
                else:
                    raise NotImplementedError

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                self.result = pa.array([int((min + max) / 2)], type=pa.int64())

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                self.result = pa.array([name_values[0][0]], pa.string())

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                self.result = pa.array([(min + max) / 2], type=pa.float64())

            def Text(self, encoding: str) -> None:
                try:
                    char_set = json.loads(self.properties[TEXT_CHARSET])
                except json.JSONDecodeError:
                    self.result = pa.array([""], pa.string())
                else:
                    max_length = int(self.properties[TEXT_MAX_LENGTH])
                    ex = ''.join(char_set)
                    if len(ex) > max_length:
                        ex = ex[:max_length]
                    self.result = pa.array([ex], pa.string())

            def Bytes(self) -> None:
                self.result = pa.array(
                    [bytes('1', 'utf-8')], pa.binary(length=1)
                )

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = pa.StructArray.from_arrays(
                    arrays=[
                        field_type.example() for field_type in fields.values()
                    ],
                    names=list(fields.keys()),
                )

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                n_fields = len(fields)
                arrays = []
                for j, field_type in enumerate(fields.values()):
                    middle_arr = field_type.example()
                    early_arr = pa.nulls(j, type=middle_arr.type)
                    late_arr = pa.nulls(n_fields - j - 1, type=middle_arr.type)
                    arrays.append(
                        pa.concat_arrays([early_arr, middle_arr, late_arr])
                    )
                names = list(fields.keys())
                arrays.append(pa.array(names, pa.string()))
                names.append('field_selected')
                self.result = pa.StructArray.from_arrays(
                    arrays=arrays,
                    names=names,
                )

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = type.example()

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Datetime(
                self, format: str, min: str, max: str, base: sdt.DatetimeBase
            ) -> None:
                self.result = pa.array(
                    pd.to_datetime([max], format=format), pa.timestamp('ns')
                )

            def Constrained(
                self,
                type: sdt.Type,
                constraint: sdt.Predicate,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

        visitor = ToArrow(properties=self.properties())
        self.accept(visitor)
        return visitor.result

    def numpy_example(self) -> np.ndarray:
        """Returns an example of numpy array matching the type.
        For an optional type, it returns a non missing value of the type.
        """
        return self.example().to_numpy(zero_copy_only=False)  # type:ignore

    def tensorflow_example(self) -> t.Any:
        """This methods returns a dictionary where the leaves are
        tf tensors. For optional types, we consider the case
        where the field is not missing.
        """

        class TensorflowExample(sdt.TypeVisitor):
            result = {}

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties if properties is not None else {}

            def Null(self) -> None:
                raise NotImplementedError

            def Unit(self) -> None:
                self.result = tf.constant([np.NaN], dtype=tf.float64)

            def Boolean(self) -> None:
                self.result = tf.constant([1], dtype=tf.int64)

            def Id(
                self,
                unique: bool,
                reference: t.Optional[sdt.Path] = None,
                base: t.Optional[sdt.IdBase] = None,
            ) -> None:
                # TODO: we should clarify for Ids, user_input
                # and so on, to be consistent
                if base == sdt.IdBase.STRING:
                    self.result = tf.constant(['1'], tf.string)
                elif base == sdt.IdBase.INT64:
                    self.result = tf.constant([1], tf.int64)
                else:
                    raise NotImplementedError

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                self.result = tf.constant([int((min + max) / 2)], tf.int64)

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                self.result = tf.constant([name_values[0][0]], tf.string)

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                self.result = tf.constant([(min + max) / 2], dtype=tf.float64)

            def Text(self, encoding: str) -> None:
                try:
                    char_set = json.loads(self.properties[TEXT_CHARSET])
                except json.JSONDecodeError:
                    self.result = tf.constant([""], tf.string)
                else:
                    max_length = int(self.properties[TEXT_MAX_LENGTH])
                    ex = ''.join(char_set)
                    if len(ex) > max_length:
                        ex = ex[:max_length]
                    self.result = tf.constant([ex], tf.string)

            def Bytes(self) -> None:
                self.result = tf.constant(['1'], tf.string)

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = {
                    field_name: field_type.tensorflow_example()
                    for field_name, field_type in fields.items()
                }

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = {
                    field_name: field_type.tensorflow_example()
                    for field_name, field_type in fields.items()
                }

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = {
                    'input_mask': tf.constant([1], dtype=tf.int64),
                    'values': type.tensorflow_example(),
                }

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Datetime(
                self, format: str, min: str, max: str, base: sdt.DatetimeBase
            ) -> None:
                self.result = tf.constant([min], dtype=tf.string)

            def Constrained(
                self,
                type: sdt.Type,
                constraint: sdt.Predicate,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

        visitor = TensorflowExample(properties=self.properties())
        self.accept(visitor)
        return visitor.result

    def default(self) -> pa.Array:
        """This methods returns a pyarrow scalar that matches the type.
        For an optional type, we consider the case where, the field
        is missing.
        """

        class Default(sdt.TypeVisitor):
            result = pa.nulls(0)

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties if properties is not None else {}

            def Null(self) -> None:
                raise NotImplementedError

            def Unit(self) -> None:
                self.result = pa.nulls(0)

            def Boolean(self) -> None:
                self.result = pa.array([True], type=pa.bool_())

            def Id(
                self,
                unique: bool,
                reference: t.Optional[sdt.Path] = None,
                base: t.Optional[sdt.IdBase] = None,
            ) -> None:
                # TODO: we should clarify for Ids, user_input
                # and so on, to be consistent
                if base == sdt.IdBase.STRING:
                    self.result = pa.array(['1'], pa.string())
                elif base == sdt.IdBase.INT64:
                    self.result = pa.array([1], pa.int64())
                else:
                    raise NotImplementedError

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                self.result = pa.array([int((min + max) / 2)], type=pa.int64())

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                self.result = pa.array([name_values[0][0]], pa.string())

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                self.result = pa.array([(min + max) / 2], type=pa.float64())

            def Text(self, encoding: str) -> None:
                try:
                    char_set = json.loads(self.properties[TEXT_CHARSET])
                except json.JSONDecodeError:
                    self.result = pa.array([""], pa.string())
                else:
                    max_length = int(self.properties[TEXT_MAX_LENGTH])
                    ex = ''.join(char_set)
                    if len(ex) > max_length:
                        ex = ex[:max_length]
                    self.result = pa.array([ex], pa.string())

            def Bytes(self) -> None:
                self.result = pa.array(
                    [bytes('1', 'utf-8')], pa.binary(length=1)
                )

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = pa.StructArray.from_arrays(
                    arrays=[
                        field_type.default() for field_type in fields.values()
                    ],
                    names=list(fields.keys()),
                )

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                n_fields = len(fields)
                arrays = []
                for j, field_type in enumerate(fields.values()):
                    middle_arr = field_type.default()
                    early_arr = pa.nulls(j, type=middle_arr.type)
                    late_arr = pa.nulls(n_fields - j - 1, type=middle_arr.type)
                    arrays.append(
                        pa.concat_arrays([early_arr, middle_arr, late_arr])
                    )
                names = list(fields.keys())
                arrays.append(pa.array(names, pa.string()))
                names.append('field_selected')
                self.result = pa.StructArray.from_arrays(
                    arrays=arrays,
                    names=names,
                )

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = pa.array([None], type=type.default().type)

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Datetime(
                self, format: str, min: str, max: str, base: sdt.DatetimeBase
            ) -> None:
                self.result = pa.array(
                    pd.to_datetime([max], format=format), pa.timestamp('ns')
                )

            def Constrained(
                self,
                type: sdt.Type,
                constraint: sdt.Predicate,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

        visitor = Default(properties=self.properties())
        self.accept(visitor)
        return visitor.result

    def numpy_default(self) -> np.ndarray:
        """Returns an example of numpy array matching the type.
        For an optional type, it sets the default missing value
        """
        return self.default().to_numpy(zero_copy_only=False)  # type:ignore

    def tensorflow_default(self, is_optional: bool = False) -> t.Any:
        """This methods returns a dictionary with tensors as leaves
        that match the type.
        For an optional type, we consider the case where the field
        is missing, and set the default value for each missing type.
        """

        class ToTensorflow(sdt.TypeVisitor):
            result = {}

            def __init__(
                self, properties: t.Optional[t.Mapping[str, str]] = None
            ):
                self.properties = properties if properties is not None else {}

            def Null(self) -> None:
                raise NotImplementedError

            def Unit(self) -> None:
                self.result = tf.constant([np.NaN], dtype=tf.float64)

            def Boolean(self) -> None:
                if is_optional:
                    self.result = tf.constant(
                        [np.iinfo(np.int64).max], dtype=tf.int64
                    )
                else:
                    self.result = tf.constant([1], dtype=tf.int64)

            def Id(
                self,
                unique: bool,
                reference: t.Optional[sdt.Path] = None,
                base: t.Optional[sdt.IdBase] = None,
            ) -> None:
                # TODO: we should clarify for Ids, user_input
                # and so on, to be consistent
                if is_optional:
                    if base == sdt.IdBase.STRING:
                        self.result = tf.constant([''], dtype=tf.string)
                    elif base == sdt.IdBase.INT64:
                        self.result = tf.constant(
                            [np.iinfo(np.int64).max], pa.string()
                        )
                    else:
                        raise NotImplementedError
                else:
                    if base == sdt.IdBase.STRING:
                        self.result = tf.constant(['1'], tf.string)
                    elif base == sdt.IdBase.INT64:
                        self.result = tf.constant([1], tf.int64)
                    else:
                        raise NotImplementedError

            def Integer(
                self, min: int, max: int, base: sdt.IntegerBase
            ) -> None:
                if is_optional:
                    self.result = tf.constant(
                        [np.iinfo(np.int64).min], dtype=tf.int64
                    )
                else:
                    self.result = tf.constant(
                        [int((min + max) / 2)], type=tf.int64
                    )

            def Enum(
                self,
                name: str,
                name_values: t.Sequence[t.Tuple[str, int]],
                ordered: bool,
            ) -> None:
                if is_optional:
                    self.result = tf.constant([''], dtype=tf.string)
                else:
                    self.result = tf.constant([name_values[0][0]], tf.string)

            def Float(
                self, min: float, max: float, base: sdt.FloatBase
            ) -> None:
                if is_optional:
                    self.result = tf.constant([np.NaN], dtype=tf.float64)
                else:
                    self.result = tf.constant(
                        [(min + max) / 2], dtype=tf.float64
                    )

            def Text(self, encoding: str) -> None:
                if is_optional:
                    self.result = tf.constant([''], dtype=tf.string)
                else:
                    try:
                        char_set = json.loads(self.properties[TEXT_CHARSET])
                    except json.JSONDecodeError:
                        self.result = tf.constant([""], tf.string)
                    else:
                        max_length = int(self.properties[TEXT_MAX_LENGTH])
                        ex = ''.join(char_set)
                        if len(ex) > max_length:
                            ex = ex[:max_length]
                        self.result = tf.constant([ex], tf.string)

            def Bytes(self) -> None:
                if is_optional:
                    self.result = tf.constant([''], dtype=tf.string)
                else:
                    self.result = tf.constant(['1'], tf.string)

            def Struct(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = {
                    field_name: field_type.tensorflow_default(
                        is_optional=is_optional
                    )
                    for field_name, field_type in fields.items()
                }

            def Union(
                self,
                fields: t.Mapping[str, sdt.Type],
                name: t.Optional[str] = None,
            ) -> None:
                self.result = {
                    field_name: field_type.tensorflow_default(
                        is_optional=is_optional
                    )
                    for field_name, field_type in fields.items()
                }

            def Optional(
                self, type: sdt.Type, name: t.Optional[str] = None
            ) -> None:
                self.result = {
                    'input_mask': tf.constant([0], dtype=tf.int64),
                    'values': type.tensorflow_default(is_optional=True),
                }

            def List(
                self,
                type: sdt.Type,
                max_size: int,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Array(
                self,
                type: sdt.Type,
                shape: t.Tuple[int, ...],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Datetime(
                self, format: str, min: str, max: str, base: sdt.DatetimeBase
            ) -> None:
                if is_optional:
                    self.result = tf.constant([''], dtype=tf.string)
                else:
                    self.result = tf.constant([min], dtype=tf.string)

            def Constrained(
                self,
                type: sdt.Type,
                constraint: sdt.Predicate,
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

            def Hypothesis(
                self,
                *types: t.Tuple[sdt.Type, float],
                name: t.Optional[str] = None,
            ) -> None:
                raise NotImplementedError

        visitor = ToTensorflow(properties=self.properties())
        self.accept(visitor)
        return visitor.result


# A few builders
def Null() -> Type:
    return Type(sp.Type(name='Null', null=sp.Type.Null()))


def Unit(properties: t.Optional[t.Mapping[str, str]] = None) -> Type:
    return Type(
        sp.Type(name='Unit', unit=sp.Type.Unit(), properties=properties)
    )


def Id(
    unique: bool,
    base: t.Optional[sdt.IdBase] = None,
    reference: t.Optional[sdt.Path] = None,
) -> Type:
    if base is None:
        base = sdt.IdBase.STRING
    if reference is None:
        return Type(
            sp.Type(name='Id', id=sp.Type.Id(base=base.value, unique=unique))
        )
    return Type(
        sp.Type(
            name='Id',
            id=sp.Type.Id(
                base=base.value, unique=unique, reference=reference.protobuf()
            ),
        )
    )


def Boolean() -> Type:
    return Type(sp.Type(name='Boolean', boolean=sp.Type.Boolean()))


def Integer(
    min: t.Optional[int] = None,
    max: t.Optional[int] = None,
    base: t.Optional[sdt.IntegerBase] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    if base is None:
        base = sdt.IntegerBase.INT64
    if min is None:
        if base == sdt.IntegerBase.INT64:
            min = np.iinfo(np.int64).min
        elif base == sdt.IntegerBase.INT32:
            min = np.iinfo(np.int32).min
        elif base == sdt.IntegerBase.INT16:
            min = np.iinfo(np.int16).min
        else:
            min = np.iinfo(np.int8).min
    if max is None:
        if base == sdt.IntegerBase.INT64:
            max = np.iinfo(np.int64).max
        elif base == sdt.IntegerBase.INT32:
            max = np.iinfo(np.int32).max
        elif base == sdt.IntegerBase.INT16:
            max = np.iinfo(np.int16).max
        else:
            max = np.iinfo(np.int8).max
    return Type(
        sp.Type(
            name='Integer',
            integer=sp.Type.Integer(base=base.value, min=min, max=max),
            properties=properties,
        )
    )


def Enum(
    name: str,
    name_values: t.Union[
        t.Sequence[str], t.Sequence[int], t.Sequence[t.Tuple[str, int]]
    ],
    ordered: bool = False,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    enum_name_values: t.List[sp.Type.Enum.NameValue]
    if len(name_values) == 0:
        raise ValueError("No enum values")
    if isinstance(name_values[0], str):
        name_values = t.cast(t.Sequence[str], name_values)
        enum_name_values = [
            sp.Type.Enum.NameValue(name=n, value=v)
            for v, n in enumerate(sorted(name_values))
        ]
    elif isinstance(name_values[0], int):
        name_values = t.cast(t.Sequence[int], name_values)
        enum_name_values = [
            sp.Type.Enum.NameValue(name=str(v), value=v)
            for v in sorted(name_values)
        ]
    elif isinstance(name_values[0], tuple):
        name_values = t.cast(t.Sequence[t.Tuple[str, int]], name_values)
        enum_name_values = [
            sp.Type.Enum.NameValue(name=n, value=v)
            for n, v in sorted(name_values)
        ]
    return Type(
        sp.Type(
            name=name,
            enum=sp.Type.Enum(
                base=sp.Type.Enum.Base.INT64,
                ordered=ordered,
                name_values=enum_name_values,
            ),
            properties=properties,
        )
    )


def Float(
    min: t.Optional[float] = np.finfo(np.float64).min,
    max: t.Optional[float] = np.finfo(np.float64).max,
    base: t.Optional[sdt.FloatBase] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    if base is None:
        base = sdt.FloatBase.FLOAT64
    if min is None:
        if base == sdt.FloatBase.FLOAT64:
            min = np.finfo(np.float64).min
        elif base == sdt.FloatBase.FLOAT32:
            min = np.finfo(np.float32).min
        else:
            min = np.finfo(np.float16).min
    if max is None:
        if base == sdt.FloatBase.FLOAT64:
            max = np.finfo(np.float64).max
        elif base == sdt.FloatBase.FLOAT32:
            max = np.finfo(np.float32).max
        else:
            max = np.finfo(np.float16).max
    return Type(
        sp.Type(
            name='Float64',
            float=sp.Type.Float(base=base.value, min=min, max=max),
            properties=properties,
        )
    )


def Text(
    encoding: str = 'UTF-8', properties: t.Optional[t.Mapping[str, str]] = None
) -> Type:
    return Type(
        sp.Type(
            name=f'Text {encoding}',
            text=sp.Type.Text(encoding='UTF-8'),
            properties=properties,
        )
    )


def Bytes() -> Type:
    return Type(sp.Type(name='Bytes', bytes=sp.Type.Bytes()))


def Struct(
    fields: t.Mapping[str, sdt.Type],
    name: str = 'Struct',
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name=name,
            struct=sp.Type.Struct(
                fields=[
                    sp.Type.Struct.Field(name=name, type=type.protobuf())
                    for name, type in fields.items()
                ]
            ),
            properties=properties,
        )
    )


def Union(
    fields: t.Mapping[str, sdt.Type],
    name: str = 'Union',
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name=name,
            union=sp.Type.Union(
                fields=[
                    sp.Type.Union.Field(
                        name=field_name, type=field_type.protobuf()
                    )
                    for field_name, field_type in fields.items()
                ]
            ),
            properties=properties,
        )
    )


def Optional(
    type: sdt.Type,
    name: str = 'Optional',
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name=name,
            optional=sp.Type.Optional(type=type.protobuf()),
            properties=properties,
        )
    )


def List(
    type: sdt.Type,
    max_size: int = np.iinfo(np.int64).max,
    name: str = 'List',
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name=name,
            list=sp.Type.List(type=type.protobuf(), max_size=max_size),
            properties=properties,
        )
    )


def Array(
    type: sdt.Type,
    shape: t.Sequence[int],
    name: str = 'Array',
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name=name,
            array=sp.Type.Array(type=type.protobuf(), shape=shape),
            properties=properties,
        )
    )


def Datetime(
    format: t.Optional[str] = None,
    min: t.Optional[str] = None,
    max: t.Optional[str] = None,
    base: t.Optional[sdt.DatetimeBase] = None,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    if format is None:
        format = '%Y-%m-%dT%H:%M:%S'
    if base is None:
        base = sdt.DatetimeBase.INT64_NS
    assert base == sdt.DatetimeBase.INT64_NS
    bounds = []
    iint64 = np.iinfo(np.int64)
    for i, bound in enumerate((min, max)):
        if bound is None:  # the bound is assumed to be sound otherwise
            # This starts with the true bounds for the type datetime64[ns]
            # However, storing dates as string implies an aliasing:
            # datetime.datetime cannot be more precise than µs.
            # So this would truncate the nanoseconds:
            # ```
            # min = (min + np.timedelta64(1, "us")).astype("datetime64[us]")
            # max = max.astype("datetime64[us]")
            # ```
            # More generally, the date format can only store a bound lower
            # than that bound, which is fine with the max but not for the
            # min, as it truncates some time units.
            if i == 0:
                bound = iint64.min + 1  # iint64.min maps to 'NaT'
                aliasing = np.timedelta64(0, 'ns')
                # This looks for the lowest offset for the format:
                # see:
                # https://numpy.org/doc/stable/reference/arrays.datetime.html#datetime-and-timedelta-arithmetic
                for unit, np_unit in [
                    ("%Y", "Y"),
                    ("%m", "M"),
                    ("%d", "D"),
                    ("%H", "h"),
                    ("%M", "m"),
                    ("%S", "s"),
                    ("%f", "us"),
                ]:
                    if unit not in format:
                        break
                    elif unit in ["%m", "%Y"]:
                        # months and years have variable length
                        as_unit = np.datetime64(bound, np_unit)
                        aliasing = np.timedelta64(1, np_unit)
                        aliasing = as_unit + aliasing - as_unit
                        aliasing = aliasing.astype("timedelta64[ns]")
                    else:
                        aliasing = np.timedelta64(1, np_unit)
            elif i == 1:
                bound = iint64.max
                aliasing = np.timedelta64(0, 'ns')
            bound = str(
                (np.datetime64(bound, "ns") + aliasing).astype(
                    "datetime64[us]"
                )
            )
            bound = datetime.datetime.strptime(
                bound, '%Y-%m-%dT%H:%M:%S.%f'
            ).strftime(format)
        bounds.append(bound)

    return Type(
        sp.Type(
            name='Datetime',
            datetime=sp.Type.Datetime(
                format=format, min=bounds[0], max=bounds[1], base=base.value
            ),
            properties=properties,
        )
    )


def Constrained(
    type: sdt.Type,
    constraint: Predicate,
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name='Constrained',
            constrained=sp.Type.Constrained(
                type=type.protobuf(), constraint=constraint._protobuf
            ),
            properties=properties,
        )
    )


def Hypothesis(
    *types: t.Tuple[sdt.Type, float],
    name: str = 'Hypothesis',
    properties: t.Optional[t.Mapping[str, str]] = None,
) -> Type:
    return Type(
        sp.Type(
            name=name,
            hypothesis=sp.Type.Hypothesis(
                types=(
                    sp.Type.Hypothesis.Scored(type=v.protobuf(), score=s)
                    for v, s in types
                )
            ),
            properties=properties,
        )
    )


def extract_filter_from_types(
    initial_type: sdt.Type, goal_type: sdt.Type
) -> sdt.Type:
    class FilterVisitor(sdt.TypeVisitor):
        """Visitor that select type for filtering, it only takes
        the Union types of the goal type and the rest is taken from
        the initial type
        """

        filter_type = initial_type

        def Union(
            self,
            fields: t.Mapping[str, sdt.Type],
            name: t.Optional[str] = None,
        ) -> None:
            # here select the fields in the goal type
            self.filter_type = Union(
                fields={
                    field_name: extract_filter_from_types(
                        initial_type=initial_type.children()[field_name],
                        goal_type=field_type,
                    )
                    for field_name, field_type in fields.items()
                }
            )

        def Struct(
            self,
            fields: t.Mapping[str, sdt.Type],
            name: t.Optional[str] = None,
        ) -> None:
            # here select the fields in the initial type
            self.filter_type = Struct(
                fields={
                    field_name: (
                        extract_filter_from_types(
                            initial_type=field_type,
                            goal_type=fields[field_name],
                        )
                        if fields.get(field_name) is not None
                        else field_type
                    )
                    for field_name, field_type in initial_type.children().items()  # noqa: E501
                }
            )

        def Optional(
            self, type: sdt.Type, name: t.Optional[str] = None
        ) -> None:
            # here it does not change
            self.filter_type = Optional(
                type=extract_filter_from_types(
                    initial_type=initial_type.children()[OPTIONAL_VALUE],
                    goal_type=type,
                )
            )

    visitor = FilterVisitor()
    goal_type.accept(visitor)
    return visitor.filter_type


def extract_project_from_types(
    initial_type: sdt.Type, goal_type: sdt.Type
) -> sdt.Type:
    class ProjectVisitor(sdt.TypeVisitor):
        """Visitor that select type for projecting, it only takes
        the Project types of the goal type and the rest is taken from
        the initial type
        """

        project_type = initial_type

        def Union(
            self,
            fields: t.Mapping[str, sdt.Type],
            name: t.Optional[str] = None,
        ) -> None:
            # here select the fields in the initial type
            self.project_type = Union(
                fields={
                    field_name: (
                        extract_filter_from_types(
                            initial_type=field_type,
                            goal_type=fields[field_name],
                        )
                        if fields.get(field_name) is not None
                        else field_type
                    )
                    for field_name, field_type in initial_type.children().items()  # noqa: E501
                }
            )

        def Struct(
            self,
            fields: t.Mapping[str, sdt.Type],
            name: t.Optional[str] = None,
        ) -> None:
            # here select the fields in the goal type
            self.project_type = Struct(
                fields={
                    field_name: extract_project_from_types(
                        initial_type=initial_type.children()[field_name],
                        goal_type=field_type,
                    )
                    for field_name, field_type in fields.items()
                }
            )

        def Optional(
            self, type: sdt.Type, name: t.Optional[str] = None
        ) -> None:
            # here it does not change
            self.project_type = Optional(
                type=extract_filter_from_types(
                    initial_type=initial_type.children()[OPTIONAL_VALUE],
                    goal_type=type,
                )
            )

    visitor = ProjectVisitor()
    goal_type.accept(visitor)
    return visitor.project_type


if t.TYPE_CHECKING:
    test_type: sdt.Type = Type(sp.Type())



import typing as t

import pyarrow as pa

from sarus_data_spec.type import Type
import sarus_data_spec.arrow.type as arrow_type
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


def to_arrow(schema: sp.Schema) -> pa.schema:
    """Convert Sarus schema to pyarrow schema."""
    return arrow_schema(Type(schema.type))


def arrow_schema(_type: st.Type) -> pa.Schema:
    """Visitor that returns the schema Arrow given the Sarus Type
    #TODO: Currently only Struct and Unions are supported
    """

    class SchemaVisitor(st.TypeVisitor):
        schema: pa.Schema = pa.schema(fields=[])

        def Integer(self, min: int, max: int, base: st.IntegerBase) -> None:
            pass

        def Float(self, min: float, max: float, base: st.FloatBase) -> None:
            pass

        def Datetime(
            self,
            format: str,
            min: str,
            max: str,
            base: st.DatetimeBase,
        ) -> None:
            pass

        def Struct(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
        ) -> None:
            self.schema = pa.schema(
                fields=[
                    pa.field(
                        name=field_name,
                        type=arrow_type.to_arrow(
                            fields[field_name].protobuf()
                        ),
                    )
                    for field_name in fields.keys()
                ]
            )

        def Constrained(
            self,
            type: st.Type,
            constraint: st.Predicate,
            name: t.Optional[str] = None,
        ) -> None:
            pass

        def List(
            self,
            type: st.Type,
            max_size: int,
            name: t.Optional[str] = None,
        ) -> None:
            # TODO
            pass

        def Array(
            self,
            type: st.Type,
            shape: t.Tuple[int, ...],
            name: t.Optional[str] = None,
        ) -> None:
            # TODO
            pass

        def Optional(
            self, type: st.Type, name: t.Optional[str] = None
        ) -> None:
            pass

        def Union(
            self,
            fields: t.Mapping[str, st.Type],
            name: t.Optional[str] = None,
        ) -> None:
            arrow_fields = [
                pa.field(
                    name=field_name,
                    type=arrow_type.to_arrow(field_type.protobuf()),
                )
                for field_name, field_type in fields.items()
            ]
            arrow_fields.append(
                pa.field(name='field_selected', type=pa.string())
            )
            self.schema = pa.schema(fields=arrow_fields)

    visitor = SchemaVisitor()
    _type.accept(visitor)
    return visitor.schema
